#include <stdio.h>
#include <math.h>
#include <sys/file.h>
#include <sys/param.h>
#include <sys/types.h>
#include <unistd.h>
#include <xview/xview.h>
#include <xview/textsw.h>
#include "bt_ui.h"
#include <ctype.h>
#include "bt.h"

/* local globals */
char            value[BUFSIZ], buffer[BUFSIZ], entry[BUFSIZ], key[BUFSIZ],
                label[BUFSIZ];
int             count, nextout, extension;
char            obracket, cbracket;
char            value_obracket, value_cbracket;
extern char    *fields[];
extern int      getbuf();

#define KEY_OK 101
#define VALUE_OK 102
#define END_OF_ITEM 103
#define FIELD_OK 104

int
getch()
{				/* returns next char or EOF */
	if (count > 0) {
		count--;
		return buffer[nextout++];
	} else if (count == 0) {
		return getbuf();
	} else
		return EOF;
}


int
getbuf()
{				/* fills buffer, sets count & nextout and
				 * returns 1st char or EOF */
	count = fread(buffer, 1, BUFSIZ, infile);
	if (count == 0)
		return EOF;

	nextout = 0;
	return getch();

}



int
scan_to(target_chr)
	int             target_chr;

{
	int             ch;
	ch = getch();
	while ((ch != target_chr) && (ch != EOF))
		ch = getch();
	return ch;
}

int
get_entry_type()
{
	int             ch, wcount = 0;

	obracket = '\0';

	if ((ch = getch()) == EOF)
		return EOF;
	while (!isgraph(ch)) {
		if ((ch = getch()) == EOF)
			return EOF;
	}
	entry[wcount++] = ch;
	if ((ch = getch()) == EOF)
		return EOF;
	while (isgraph(ch)) {
		if (open_bracket(ch)) {
			obracket = ch;
			break;
		}
		entry[wcount++] = ch;
		if ((ch = getch()) == EOF)
			return EOF;
	}
	entry[wcount++] = '\0';

	return ch;
}

int
get_key()
{
	int             ch, wcount = 0;

	if ((ch = getch()) == EOF)
		return EOF;
	while (!isgraph(ch)) {
		if ((ch = getch()) == EOF)
			return EOF;
	}
	key[wcount++] = ch;
	if ((ch = getch()) == EOF)
		return EOF;
	while (isgraph(ch)) {
		if (ch == ',') {
			break;
		}
		key[wcount++] = ch;
		if ((ch = getch()) == EOF)
			return EOF;
	}
	if (ch != ',')
		scan_to(',');
	key[wcount++] = '\0';
	return KEY_OK;
}
int
close_bracket(ch)
	int             ch;
{
	return (ch == cbracket);/* returns true if ch is the current close
				 * bracket */
}
int
get_label()
{
	int             ch, wcount = 0;

	if ((ch = getch()) == EOF)
		return EOF;
	while (!isgraph(ch)) {
		if ((ch = getch()) == EOF)
			return EOF;
	}
	if (ch == cbracket) {
		return ch;
	}
	label[wcount++] = ch;
	if ((ch = getch()) == EOF) {
		return EOF;
	}
	while (isgraph(ch)) {
		if (ch == '=') {
			break;
		}
		label[wcount++] = ch;
		if ((ch = getch()) == EOF)
			return EOF;
	}
	label[wcount++] = '\0';
	return ch;
}

void
store_field(ref)
	Ref             ref;
{
	int             i;
	char            temp[BUFSIZ * 2];
	static	char	old_label[BUFSIZ];

	i = 1;
	while (i < 25) {
		if (i == 24) {
			if (strcasecmp(label, "annote") == 0) {
				add_to(ref->annote, value);
				ref->open_char[i] = value_obracket;
				ref->close_char[i] = value_cbracket;
			} else {
				if (ref->nsfield) {
					add_to(ref->nsfield, ",\n");
				} else {
					strcpy(old_label,"");
				}
				if (strcmp(old_label,label) == 0) {
					ref->nsfield[ strlen(ref->nsfield) - 3] = '\0';
					sprintf(temp, "%s%c",  value, value_cbracket);
				}
				else {
					sprintf(temp, "%s = %c%s%c", label,value_obracket, value, value_cbracket);
				}
				add_to(ref->nsfield, temp);
				strcpy(old_label,label);
			}
		} else if (strcasecmp(label, fields[i]) == 0) {
			add_to(ref->field[i], value);
			if (extension) {    		/* note the extension */
				ref->open_char[i] = ABBREVCODE;
			} else {	
				ref->open_char[i] = value_obracket;
				ref->close_char[i] = value_cbracket;
			}
			break;
		}
		i++;
	}			/* end of while  */
}

int
get_value()
{
	int             ch, i, wcount = 0;
	int             lparn, rparn, abrev, quote, ccount;
	char			temp[BUFSIZ],temp2[BUFSIZ];

	extension = FALSE;
	lparn = rparn = abrev = quote = 0;
	if ((ch = getch()) == EOF)
		return EOF;
	while (!isgraph(ch)) {
		if ((ch = getch()) == EOF)
			return EOF;
	}
	while (1) {		/* loop until the value terminator has been
				 * found and we have returned */
		if (open_bracket(ch)) {
			value_obracket = ch;
			value_cbracket = match_bracket(ch);
			lparn = 1;
			if ((ch = getch()) == EOF)
				return EOF;
		} else if (ch == '"') {
			quote = 1;
			value_obracket = '"';
			value_cbracket = '"';
			if ((ch = getch()) == EOF)
				return EOF;
		} else
			abrev = 1;
		while (1) {
			if (lparn) {
				if (ch == value_cbracket)
					rparn++;
				else if (ch == value_obracket)
					lparn++;
				if (lparn == rparn)
					break;
			} else if (quote) {
				if (value[wcount-1] != '\\') { /* skip backslash quote */
				if (ch == '"')
					break;
				}
			} else {
				if ((ch == ',') || (ch == cbracket))
					break;
			}
	/*		if ( iscntrl(ch) )   /* all control characters are treated
									as a space-- is this ok?
									
				if (strcasecmp(label, "annote") != 0)
					ch = ' '; */
			value[wcount++] = ch;
			if (wcount == BUFSIZ - 1) {
				value[wcount] = '\0';
				store_field(ref);
				wcount = 0;
			}
			if ((ch = getch()) == EOF)
				return EOF;
		}
		value[wcount] = '\0';
		if ((abrev) && (!extension) && (!isNumber(value))) {
			sprintf(temp, "%c%s", ABBREVCODE, value);
			strcpy(value, temp);
			wcount++;
		}
		/* look for comma, concat char, or close bracket */
		ccount = 0;
		if (lparn) { /* read the next character so the test for close bracket
					will not succeed */
			temp[ccount++] = ch;
			if ((ch = getch()) == EOF)
				return EOF;
		}
		while ((ch != ',') && (ch != cbracket) && (ch != '#')) {
			temp[ccount++] = ch;
			if ((ch = getch()) == EOF)
				return EOF;
		}
		if (ch == cbracket) {
			return END_OF_ITEM;
		} else if (ch == ',') {
			return FIELD_OK;

		} else if (ch == '#') {	/* it has to be this */
			extension = TRUE;
			if (quote) 				/* add back the open bracket character */
				value_obracket = '"';
			sprintf(temp2, "%c%s", value_obracket, value);
			strcpy(value, temp2);
			wcount++;
									/* add the intervening stuff */
			temp[ccount++] = ch;
			for (i = 0; i < ccount; i++) {
				value[wcount++] = temp[i];
				if (wcount == BUFSIZ - 1) {
					value[wcount] = '\0';
					store_field(ref);
					wcount = 0;
				}
			}
			/*
			 * and don't return but get next value
			 */

			lparn = rparn = abrev = quote = 0;
			if ((ch = getch()) == EOF)
				return EOF;
			while (!isgraph(ch)) {
				value[wcount++] = ch;
				if (wcount == BUFSIZ - 1) {
					value[wcount] = '\0';
					store_field(ref);
					wcount = 0;
				}
				if ((ch = getch()) == EOF)
					return EOF;
			}
		}
	}
}

int
get_field(ref)
	Ref             ref;
{
	int             result;

	if ((result = get_label()) == EOF)
		return EOF;
	if (close_bracket(result))
		return END_OF_ITEM;

	if (result != '=') {

		if ((result = scan_to('=')) == EOF)
			return EOF;
	}
	if ((result = get_value()) == EOF)
		return EOF;

	store_field(ref);
	return result;
}

int
get_item()
{
	int             result;

	if ((result = get_key()) == EOF)
		return EOF;

	if ((ref = (Ref) malloc(sizeof(Reference))) == NULL) {
		fprintf(stderr, "bibcard: no memory for load\n");
		exit(0);
	}
	bzero((char *) ref, sizeof(Reference));

	strcpy(ref->ref_type, entry);
	toLower(ref->ref_type);
	set_str(ref->field[0], key);
	if (duplicate_cite_key(key))
		fprintf(stderr, "bibcard: duplicate cite key: %s\n", key);
	if (first_item == 0) {
		first_item = ref;
		prev = ref;
	} else {
		ref->next = prev->next;
		ref->prev = prev;
		prev->next = ref;
		prev = ref;
	}
	inum++;
	result = get_field(ref);
	while (result == FIELD_OK) {
		result = get_field(ref);
	}
	return result;
}

int
open_bracket(ch)
	int             ch;
{
	return strchr("{[(<", ch);
}

int
match_bracket(ch)
	int             ch;

{
	switch (ch) {

	case '{':
		return '}';
		break;
	case '[':
		return ']';
		break;
	case '(':
		return ')';
		break;
	case '<':
		return '>';
		break;
	}
}

int
set_brackets()
{
	int             ch;

	if (obracket == '\0') {

		if ((ch = getch()) == EOF)
			return EOF;
		while (!isgraph(ch)) {
			if ((ch = getch()) == EOF)
				return EOF;
		}
		if (open_bracket(ch)) {
			obracket = ch;

		} else {	/* something is wrong */
			fprintf(stderr, "bibcard: no open bracket for: %s\n", entry);
			return EOF;
		}
	}
	cbracket = match_bracket(obracket);
	return cbracket;
}
void
add_to_string(temp)
	char           *temp;
{
	if (edit_count > 0)
		textsw_insert(bt_stringpop->text, "\n", 1);
	textsw_insert(bt_stringpop->text, temp, strlen(temp));
	edit_count = (int) xv_get(bt_stringpop->text, TEXTSW_EDIT_COUNT);
}

int
get_string()
{
	int             ch, wcount = 0;
	int             lparn, rparn;
	char            temp[BUFSIZ];


	if ((ch = getch()) == EOF)
		return EOF;
	lparn = 1;
	rparn = 0;

	while (lparn != rparn) {
		temp[wcount++] = ch;
		if (wcount == BUFSIZ - 1) {
			temp[wcount] = '\0';
			add_to_string(temp);
			wcount = 0;
		}
		if (ch == cbracket)
			rparn++;
		else if (ch == obracket)
			lparn++;

		if ((ch = getch()) == EOF)
			return EOF;
	}
	temp[wcount-1] = '\0';
	add_to_string(temp);
	return 0;
}

int
get_preamble()
{
	int             ch, wcount = 0;
	int             lparn, rparn;
	char            temp[BUFSIZ];


	if ((ch = getch()) == EOF)
		return EOF;
	lparn = 1;
	rparn = 0;

	while (lparn != rparn) {
		temp[wcount++] = ch;
		if (wcount == BUFSIZ - 1) {
			temp[wcount] = '\0';
			add_to(preamble, temp);
			wcount = 0;
		}
		if (ch == cbracket)
			rparn++;
		else if (ch == obracket)
			lparn++;

		if ((ch = getch()) == EOF)
			return EOF;
	}
	temp[wcount-1] = '\0';
	add_to(preamble, temp);
	return 0;
}


int
load_bib()
{

	int             ch;

	while (1) {
		obracket = '\0';
		if ((ch = scan_to('@')) == EOF)	/* scans to next @ */
			return 0;
		if ((ch = get_entry_type()) == EOF)
			return LD_ERROR;
		if ((ch = set_brackets()) == EOF)
			return LD_ERROR;

		/* we have a @entry */

		if (strcasecmp(entry, "string") == 0) {
			if ((ch = get_string()) == EOF)
				return LD_ERROR;
		} else if (strcasecmp(entry, "preamble") == 0) {
			if ((ch = get_preamble()) == EOF)
				return LD_ERROR;
		} else {	/* assume it is a new item */
			if ((ch = get_item()) == EOF)
				return LD_ERROR;
		}
	}
}
